"""Handler for "help" REPL command."""

from __future__ import annotations

from argparse import Namespace

from rich.box import HORIZONTALS
from rich.table import Table

from .__common__ import ReplCommand


# ------------------------------------------------------------------------------
@ReplCommand.register('help', 'h', '?', history=False)
class Help(ReplCommand):
    """Command help."""

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return r'h\[elp]', 'Show this help message.'

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Print help on all commands."""

        table = Table(show_header=True, leading=1, box=HORIZONTALS)
        table.add_column('Command', no_wrap=True, vertical='top')
        table.add_column('Description', vertical='top')
        for _, handler in sorted(ReplCommand.commands.items()):
            table.add_row(*handler.help())
        self.print(table)
        self.print()
        self.print(fr'Run commands as {self.cmd_escape}cmd \[options]')

        return True



"""Handler for "results" REPL command."""

from __future__ import annotations

from argparse import Namespace

from .__common__ import ReplCommand, ReplCommandError, Style


# ------------------------------------------------------------------------------
@ReplCommand.register('results', 'r')
class SetNumberOfSearchResults(ReplCommand):
    """Set the number of search results from the database to feed to the LLM."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument('n_results', nargs='?')

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return (
            fr'r\[esults] [[{Style.arg}]N]',
            'Set the number of search results from the database to feed to the LLM.',
        )

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Set the number of search results from the database to feed to the LLM."""

        if args.n_results is None:
            self.print(f'Number of search results = {self.repl_context.search_results}')
            return True

        try:
            n_results = int(args.n_results)
        except ValueError as e:
            raise ReplCommandError(f'{args.n_results}: {e}')

        if n_results <= 0:
            raise ReplCommandError('Number of search results must be > 0')

        self.repl_context.search_results = n_results
        self.print(f'Number of search results set to {n_results}')
        return True



"""Handler for "llm" REPL command."""

from __future__ import annotations

from argparse import Namespace

import ollama

from .__common__ import ReplCommand, ReplCommandError, Style


# ------------------------------------------------------------------------------
@ReplCommand.register('llm', 'l')
class SetLlm(ReplCommand):
    """Set the LLM model to be used."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument('llm', nargs='*')

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return (
            fr'l\[lm] [[{Style.arg}]MODEL ...]',
            'Show the LLMs in use or set to the ones specified.',
        )

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Set the LLM model."""

        available_llms = {m.model.removesuffix(':latest') for m in ollama.list().models}

        if args.llm:
            # Update current LLM settings
            requested_llms = {m.removesuffix(':latest') for m in args.llm}
            if unavailable_llms := requested_llms - available_llms:
                raise ReplCommandError(f'Unavailable models: {", ".join(sorted(unavailable_llms))}')
            self.repl_context.llm = sorted(requested_llms)
            self.print(f'LLM(s) set to {", ".join(self.repl_context.llm)}')
            return True

        # Just report current setting
        try:
            current_llms = set(self.repl_context.llm)
        except AttributeError:
            current_llms = set()

        self.print('[bold]Available LLMs (* = selected):')
        for llm in sorted(available_llms):
            self.print(f'[{Style.active}]  * {llm}' if llm in current_llms else f'    {llm}')
        return True



#!/usr/bin/env python3

"""Ollex CLI (See https://github.com/jin-gizmo/ollex)."""

from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path

from ollex.commands import CliCommand
from ollex.version import __version__

__author__ = 'Murray Andrews'

PROG = Path(sys.argv[0]).stem


# ------------------------------------------------------------------------------
def process_cli_args() -> Namespace:
    """Process command line arguments."""

    argp = ArgumentParser(prog=PROG, description=__doc__)

    argp.add_argument(
        '-v', '--version', action='version', version=__version__, help='Show version and exit.'
    )

    # Add the sub-commonads
    subp = argp.add_subparsers(required=True)
    for cmd in sorted(CliCommand.commands.values(), key=lambda c: c.name):
        cmd(subp).add_arguments()

    args = argp.parse_args()

    try:
        args.handler.check_arguments(args)
    except ValueError as e:
        argp.error(str(e))

    return args


# ------------------------------------------------------------------------------
def main() -> int:
    """Show time."""
    try:
        args = process_cli_args()
        args.handler.execute(args)
        return 0
    except Exception as ex:
        # Uncomment for debugging
        # raise  # noqa: ERA001
        print(f'{PROG}: {ex}', file=sys.stderr)
        return 1
    except KeyboardInterrupt:
        print(f'{PROG}: Interrupt', file=sys.stderr)
        return 2


# ------------------------------------------------------------------------------
# This only gets used during dev/test. Once deployed as a package, main() gets
# imported and run directly.
if __name__ == '__main__':
    exit(main())



"""Miscellaneous utility functions."""

from __future__ import annotations

from collections.abc import Iterator


# ------------------------------------------------------------------------------
def chunk_list(ll: list, max_len: int) -> Iterator[list]:
    """Split a list into chunks with a maximum specified length."""

    for i in range(0, len(ll), max_len):
        yield ll[i : i + max_len]


# ------------------------------------------------------------------------------
def plural(s: str, n: int) -> str:
    """Pluralise a string."""

    return s if n == 1 else f'{s}s'



"""Handler for "set" REPL command."""

from __future__ import annotations

from argparse import Namespace

from rich.box import HORIZONTALS
from rich.table import Table

from ollex.lib.argparse import StoreNameValuePair
from ollex.lib.model import LLM_OPTIONS
from .__common__ import ReplCommand, ReplCommandError, Style


# ------------------------------------------------------------------------------
@ReplCommand.register('set', 's')
class SetLlm(ReplCommand):
    """Show / set LLM options."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument(dest='options', action=StoreNameValuePair, nargs='*')

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return (
            fr's\[et] [[{Style.arg}]OPTION[/{Style.arg}]=[{Style.arg}]VALUE[/{Style.arg}]]',
            'Show / set the current (non-default) LLM options.',
        )

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Set LLM model options."""

        if args.options:
            for option, value in args.options.items():
                try:
                    # Cast the value to correct type
                    self.repl_context.llm_options[option] = LLM_OPTIONS[option](value)
                except KeyError:
                    raise ReplCommandError(f'{option}: Unknown option')
                except Exception as e:
                    raise ReplCommandError(f'{option}: {e}')

        table = Table(box=HORIZONTALS)
        table.add_column('Option')
        table.add_column('Value', justify='right')
        for option in sorted(LLM_OPTIONS):
            table.add_row(option, str(self.repl_context.llm_options.get(option, '')))

        self.print(table)

        return True



"""Interactive command handling for the REPL."""

from __future__ import annotations

import shlex
from abc import ABC, abstractmethod
from argparse import Namespace
from collections.abc import Callable
from readline import get_current_history_length, remove_history_item

from rich.console import Console

from ollex.lib.argparse import ReplArgparser

# Rich styles
Style = Namespace(
    arg='underline',  # Used for command arguments in help text.
    active='green',  # Used to indicate active item in a list
    prompt='italic',  # Used to print prompts
)


# ------------------------------------------------------------------------------
class ReplCommandError(Exception):
    """Raised when a REPL command fails."""

    pass


# ------------------------------------------------------------------------------
class ReplCommand(ABC):
    """Repl command handler."""

    commands: dict[str, type[ReplCommand]] = {}
    aliases: dict[str, str] = {}
    name = None  # Set by @register decorator for subclasses.
    help_ = None  # Set by @register from first line of docstring.
    add_to_history = True

    # --------------------------------------------------------------------------
    @classmethod
    def register(cls, name: str, *aliases: str, history: bool = True) -> Callable:
        """
        Register a REPL command handler class.

        :param name:    The name of the REPL command. Usually a long form.
        :param aliases: Aliases for the command. Usually short forms.
        :param history: Whether invocations of the command are added to command
                        history.

        This is a decorator. Usage is:

        .. code-block:: python

            @ReplCommand.register('my_command')
            class MyCommand(ReplCommand):
                ...
        """

        def decorate(cmd: type[ReplCommand]):
            """Register the command handler class."""
            cmd.name = name
            cls.commands[name] = cmd
            # A command is an alias for itself
            cls.aliases[name] = name
            for alias in aliases:
                cls.aliases[alias] = name
            cmd.add_to_history = history

            return cmd

        return decorate

    # --------------------------------------------------------------------------
    def __init__(self, repl_context: Namespace, console: Console, cmd_escape: str) -> None:
        """
        Initialize the command handler.

        :param repl_context:    A namespace containing context information for the
                                REPL that can be manipulated by commands.
        :param console:         A console for command output.
        :param cmd_escape:      A string to introduce commands.
        """

        self.repl_context = repl_context
        self.console = console
        self.cmd_escape = cmd_escape
        self.argp = ReplArgparser(add_help=False, error_fmt=f'Error: {{}} ({cmd_escape}h for help)')
        self.add_arguments()

    # --------------------------------------------------------------------------
    def print(self, *args, **kwargs):
        """Shortcut for printing to rich console."""

        # Disable rich random highlighting unless explicitly requested
        self.console.print(*args, **({'highlight': False} | kwargs))

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:  # noqa B027
        """Add the REPL command handler arguments (if any)."""
        pass

    # --------------------------------------------------------------------------
    @abstractmethod
    def execute(self, args: Namespace) -> bool:
        """
        Execute the REPL command with the specified arguments.

        :param args: Parsed arguments to pass to the REPL command.

        :return: True if the main REPL loop should continue, False otherwise.
        """

        raise NotImplementedError('execute')

    # --------------------------------------------------------------------------
    @staticmethod
    @abstractmethod
    def help() -> tuple[str, str]:
        """
        Provide help information for the REPL command.

        :return:    A tuple (usage, description). Each may contain rich style
                    markup.
        """

        raise NotImplementedError('help')


# ------------------------------------------------------------------------------
class ReplCommandController:
    """Controller / dispatcher for REPL commands."""

    # --------------------------------------------------------------------------
    def __init__(self, repl_context: Namespace, console: Console, cmd_escape: str) -> None:
        """
        Initialize the REPL command controller.

        :param repl_context:    A namespace containing context information for the
                                REPL that can be manipulated by commands.
        :param console:         A console for command output.
        :param cmd_escape:      A string to introduce commands.
        """

        self.repl_context = repl_context
        self.console = console
        self.cmd_escape = cmd_escape
        self.command_handlers = {
            k: v(repl_context, console, cmd_escape) for k, v in ReplCommand.commands.items()
        }

    # --------------------------------------------------------------------------
    def do_cmd(self, command: str) -> bool:
        """
        Execute the REPL command with the specified arguments.

        :param command: A shell style command to execute via one of the REPL commands.

        """

        argv = shlex.split(command.removeprefix(self.cmd_escape))
        if not argv:
            raise ReplCommandError('No command specified')

        try:
            handler = self.command_handlers[ReplCommand.aliases[argv[0]]]
        except KeyError:
            remove_history_item(get_current_history_length() - 1)
            raise ReplCommandError(f'{argv[0]}: No such command (try {self.cmd_escape}help)')

        if not handler.add_to_history:
            remove_history_item(get_current_history_length() - 1)

        args = handler.argp.parse_args(argv[1:])
        return handler.execute(args)



"""Ollex "query" command."""

import json
import readline  # noqa
import sys
from argparse import Namespace
from collections.abc import Iterator
from contextlib import suppress
from functools import lru_cache
from importlib import resources
from itertools import product
from os import isatty
from pathlib import Path
from random import choice
from time import time
from typing import Any, TextIO

import chromadb
import ollama
import yaml
from rich.align import Align
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.status import Status

from ollex.commands import CliCommand
from ollex.exceptions import OllexError
from ollex.lib.argparse import StoreNameValuePair
from ollex.lib.misc import plural
from ollex.lib.model import LLM_OPTIONS
from ollex.repl import ReplCommandController

EMBEDDING_MODEL = 'nomic-embed-text'  # Last resort only
MIN_QUERY_LENGTH = 10  # Applies to repl only
REPL_CMD_ESCAPE = '\\'
HISTORY_FILE = Path('~/.ollex_history').expanduser()
HISTORY_LENGTH = 100

# Number of docuemnts to return when searching the database.
N_SEARCH_RESULTS = 10

Style = Namespace(
    prompt='green',
    info='italic yellow',
    status='italic #008800',
    banner='red on white',
    error='red',
)

LLM_PROMPT_FILE = resources.files('ollex') / 'resources' / 'prompts.yaml'

BANNERS = [
    line
    for line in (resources.files('ollex') / 'resources' / 'banners.txt').read_text().splitlines()
    if line.strip() and not line.startswith('#')
]


# ------------------------------------------------------------------------------
@CliCommand.register('query')
class Query(CliCommand):
    """Interact with Ollama in REPL or batch mode."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument(
            'db_directory',
            help='Directory containing the Chroma DB containing embeddings.',
        )

        self.argp.add_argument(
            '-c',
            '--collection',
            action='append',
            default=[],
            help=(
                'Database collection to use. Multiple collections can be active at'
                ' the same time by repeating this option. A response is provided for'
                ' each active collection. If not specified and the database has'
                ' only one collection, that will be used. In interactive mode, the'
                ' collection(s) can also be specified using an interactive command.'
                ' A collection must be specified for any queries to be run.'
            ),
        )
        self.argp.add_argument(
            '-l',
            '--llm',
            metavar='MODEL',
            default=[],
            action='append',
            help=(
                'Model to be used for the LLM. Multiple models can be active at the'
                ' same time by repeating this option. A response is provided for'
                ' each active model. At least one model must be specified, either'
                f' as a command argument or via the {REPL_CMD_ESCAPE}llm'
                ' interactive command in order to process any queries.'
            ),
        )

        self.argp.add_argument(
            '-e',
            '--embedding',
            default=EMBEDDING_MODEL,
            help=(
                'The embedding used when creating storing documents in the database.'
                ' This is determined (in order or precedence) by'
                ' (1) the value stored in config.yaml in the database directory;'
                ' (2) the value of this argument;'
                f' (3) the default ({EMBEDDING_MODEL}).'
            ),
        )

        self.argp.add_argument(
            '-o',
            '--option',
            metavar='NAME=VALUE',
            dest='llm_options',
            action=StoreNameValuePair,
            default={},
            help=(
                'Set the specified LLM control option(s). Available options are:'
                f' {", ".join(sorted(LLM_OPTIONS.keys()))}.'
            ),
        )

        self.argp.add_argument(
            '-p',
            '--prompts',
            metavar='FILE.yaml',
            dest='prompts_file',
            default=str(LLM_PROMPT_FILE),
            help=(
                'Read prompts from the specified YAML file. Keys are the prompt IDs'
                ' and values are the prompt text. Each prompt must contain "{data}"'
                ' and "{question}" somewhere. If not specified, a default prompt is'
                f' used. Use the {REPL_CMD_ESCAPE}prompt interactive command to see'
                f' its content.'
            ),
        )

        self.argp.add_argument(
            '-n',
            type=int,
            metavar='COUNT',
            default=1,
            help='(Batch mode only) Run each query the specified number of times. Default 1.',
        )

        self.argp.add_argument(
            '-r',
            '--search-results',
            action='store',
            type=int,
            metavar='COUNT',
            default=N_SEARCH_RESULTS,
            help=(
                'Number of documents to return when searching the database.'
                f' Default is {N_SEARCH_RESULTS}.'
            ),
        )

    # --------------------------------------------------------------------------
    @staticmethod  # noqa B027
    def check_arguments(args: Namespace) -> None:
        """Validate arguments."""

        if args.llm_options:
            options = {}
            for k, v in args.llm_options.items():
                try:
                    options[k] = LLM_OPTIONS[k](v)
                except KeyError:
                    raise ValueError(f'Unknown LLM option: {k}')
                except (TypeError, ValueError) as e:
                    raise ValueError(f'Option {k}: {e}')
            args.llm_options = options

        # Validate that we have the LLMs requested
        available_llms = {m.model.removesuffix(':latest') for m in ollama.list().models}
        requested_llms = {m.removesuffix(':latest') for m in args.llm}
        if unavailable_llms := requested_llms - available_llms:
            raise ValueError(f'Unavailable models: {", ".join(sorted(unavailable_llms))}')
        args.llm = sorted(requested_llms)

    # --------------------------------------------------------------------------
    @staticmethod
    def execute(args: Namespace) -> int:
        """Execute the command."""

        # Open vector DB
        db_dirpath = Path(args.db_directory)
        if not db_dirpath.is_dir():
            raise FileNotFoundError(f'{args.db_directory} doesn\'t exist or is not a directory.')
        db_client = chromadb.PersistentClient(path=args.db_directory)

        # Load config file stored in root of DB folder. This is not a chroma construct.
        # It was added by our utility that populates the database.
        try:
            config = yaml.safe_load((db_dirpath / 'config.yaml').read_text())
        except FileNotFoundError:
            config = {}

        # Validate collections requested on command line.
        requested_collections = set(args.collection)
        available_collections = {c.name for c in db_client.list_collections()}
        if unavailable_collections := requested_collections - available_collections:
            raise OllexError(
                f'Unavailable collections: {", ".join(sorted(unavailable_collections))}'
            )

        if isatty(sys.stdin.fileno()) and isatty(sys.stdout.fileno()):
            try:
                with suppress(FileNotFoundError):
                    readline.read_history_file(HISTORY_FILE)
                repl(config, db_client, args)
            finally:
                with suppress(Exception):
                    readline.set_history_length(HISTORY_LENGTH)
                    readline.write_history_file(HISTORY_FILE)
        else:
            batch_loop(config, db_client, args)

        return 0


# ------------------------------------------------------------------------------
def get_queries_from_console(console: Console) -> Iterator[str]:
    """Prompt for queries."""

    while True:
        try:
            query = console.input(f'\n[{Style.prompt}]Query: ')
        except EOFError:
            return
        finally:
            print()
        if query:
            yield query.strip()


# ------------------------------------------------------------------------------
def get_queries_from_file(fp: TextIO) -> Iterator[str]:
    """Read queries from a file."""

    lines = []
    for line in fp:
        if not line.strip():
            if lines:
                yield ''.join(lines)
                lines = []
            continue
        lines.append(line)

    if lines:
        yield ''.join(lines)


# ------------------------------------------------------------------------------
def refresh_yaml_dict(filename: str) -> dict[str, Any]:
    """Read a dict from a YAML file, cache results and reread if it has been modified."""

    path = Path(filename)

    try:
        mtime, content = refresh_yaml_dict.cache[filename]  # noqa
        if path.stat().st_mtime <= mtime:
            return content
    except AttributeError:
        refresh_yaml_dict.cache = {}
    except KeyError:
        pass

    # Content doesn't exist or is out of date
    content = yaml.safe_load(path.read_text())
    if not isinstance(content, dict):
        raise ValueError('Must be dict')
    refresh_yaml_dict.cache[filename] = (path.stat().st_mtime, content)  # noqa
    return content


# ------------------------------------------------------------------------------
def get_collection_info(
    db_client: chromadb.ClientAPI, config: dict[str, Any]
) -> dict[str, dict[str, Any]]:
    """
    Merge information about collections from the config file with collections found in the DB.

    :param db_client:   Chroma DB client.
    :param config:      The meta config which may contain a "collections" key
                        which is a dict containing info on each collection.
    :return:            A dictionary with the collection info for each
                        collection actually present in the DB.
    """

    db_collections = db_client.list_collections()
    if not db_collections:
        raise OllexError('No collections found in database')

    available_collections = {c.name for c in db_client.list_collections()}
    return {c: config.get('collections', {}).get(c, {}) for c in available_collections}


# ------------------------------------------------------------------------------
@lru_cache(maxsize=50)
def search_collection(
    db_client: chromadb.ClientAPI,
    collection: str,
    embedding_model: str,
    query: str,
    n_results: int,
) -> list[str]:
    """
    Search a ChromaDB collection using the specified query and embedding model.

    :param db_client:       Chroma DB client.
    :param collection:      Collection name.
    :param embedding_model: Embedding model name.
    :param query:           The LLM query for which to search.
    :param n_results:       Number of results to return.
    :return:                A list of search result strings.
    """

    embedded_query = ollama.embed(model=embedding_model, input=query)
    return db_client.get_collection(collection).query(
        query_embeddings=embedded_query['embeddings'],
        n_results=n_results,
    )['documents'][0]


# ------------------------------------------------------------------------------
def repl(config: dict[str, Any], db_client: chromadb.ClientAPI, cli_args: Namespace) -> None:
    """
    Run an interactive REPL.

    :param config:      Configuration metadata read from the database directory.
    :param db_client:   Chroma DB client.
    :param cli_args:    CLI arguments namespace.
    """

    repl_context = Namespace(
        llm=cli_args.llm,
        llm_options=cli_args.llm_options,
        collection_info=get_collection_info(db_client, config),
        collections=sorted(cli_args.collection),
        search_results=cli_args.search_results,
        prompts=refresh_yaml_dict(cli_args.prompts_file),
    )
    # If the DB contains a single collection default to that.
    if not repl_context.collections and len(repl_context.collection_info) == 1:
        repl_context.collections = list(repl_context.collection_info)

    console = Console()
    repl_command_controller = ReplCommandController(repl_context, console, REPL_CMD_ESCAPE)

    if BANNERS:
        console.print()
        console.print(
            Align.center(
                Panel(choice(BANNERS), style=Style.banner, border_style=Style.banner)  # noqa S311
            )
        )
    console.print()
    selected_models = ', '.join(sorted(repl_context.llm)) or '(not set - use /l)'
    selected_collections = ', '.join(repl_context.collections) or '(not set - use /c)'
    console.print(
        f'[{Style.info}]'
        f'{plural("Model", len(repl_context.llm))} = {selected_models}'
        '    '
        f'{plural("Collections", len(repl_context.collections))} = {selected_collections}'
        '    '
        f'{plural("Prompt", len(repl_context.prompts))} = {", ".join(repl_context.prompts)}'
    )

    for query in get_queries_from_console(console):
        # Read prompts. This is reread for any changes.
        repl_context.prompts = refresh_yaml_dict(cli_args.prompts_file)
        if query.startswith(REPL_CMD_ESCAPE):
            try:
                if not repl_command_controller.do_cmd(query[1:]):
                    break
            except Exception as e:
                console.print(f'[{Style.error}]{e}')
            continue

        if not repl_context.llm:
            console.print(f'[{Style.error}]One or more LLMs must be set -- try {REPL_CMD_ESCAPE}l')
            continue

        if not repl_context.collections:
            console.print(f'[{Style.error}]A data collection must be set -- try {REPL_CMD_ESCAPE}c')
            continue

        if len(query) < MIN_QUERY_LENGTH:
            console.print(f'[{Style.error}]Query is too short')
            readline.remove_history_item(readline.get_current_history_length() - 1)
            continue

        # Important the LLM is first in the product to minimise model loading / unloading
        for llm, collection, (prompt_id, prompt_txt) in product(
            repl_context.llm, repl_context.collections, repl_context.prompts.items()
        ):
            status = Status(f'[{Style.status}]Searching {collection} ...')
            status.start()
            embedding_model = (
                config.get('collections', {}).get(collection, {}).get('embedding_model')
            ) or cli_args.embedding

            try:
                query_results = search_collection(
                    db_client,
                    collection,
                    embedding_model,
                    query,
                    n_results=repl_context.search_results,
                )
            except Exception as e:
                status.stop()
                console.print(f'[{Style.error}]Query failed: {e}')
                continue
            else:
                status.stop()

            status = Status(
                f'[{Style.status}]Model = {llm}   Collection = {collection}   Prompt = {prompt_id}'
            )
            status.start()
            ts = time()
            try:
                response = ollama.generate(
                    model=llm,
                    options=repl_context.llm_options,
                    prompt=prompt_txt.format(data=query_results, question=query),
                )
            except Exception as e:
                status.stop()
                console.print(f'[{Style.error}]{e}')
                continue
            else:
                status.stop()
            console.rule(
                f'[{Style.info}]Model = {llm}   Collection = {collection}   Prompt = {prompt_id}',
                style=Style.info,
            )

            answer = response['response']
            elapsed = round(time() - ts, 1)
            console.print(Markdown(answer))
            console.print(f'\n[{Style.info}]{elapsed} s')
            console.print()


# ------------------------------------------------------------------------------
def batch_loop(config: dict[str, Any], db_client: chromadb.ClientAPI, cli_args: Namespace) -> None:
    """
    Run a batch query loop.

    :param config:      Configuration metadata read from the database directory.
    :param db_client:   Chroma DB client.
    :param cli_args:    CLI arguments namespace.
    """

    if not cli_args.llm:
        raise OllexError('LLM(s) must be specified in batch mode')

    collection_names = cli_args.collection
    if not collection_names:
        db_collections = db_client.list_collections()
        if len(db_collections) == 1:
            collection_names = db_collections
        else:
            raise OllexError(
                'Collection must be specified in batch mode'
                ' if DB contains more than one collection'
            )
    collections = {name: db_client.get_collection(name=name) for name in collection_names}
    prompts = refresh_yaml_dict(cli_args.prompts_file)

    for query in get_queries_from_file(sys.stdin):
        for llm, collection, (prompt_id, prompt_txt) in product(
            cli_args.llm, collections, prompts.items()
        ):
            embedding_model = (
                config.get('collections', {}).get(collection, {}).get('embedding_model')
                or cli_args.embedding
            )
            query_results = search_collection(
                db_client,
                collection,
                embedding_model,
                query,
                n_results=cli_args.search_results,
            )
            for n in range(cli_args.n):
                ts = time()
                response = ollama.generate(
                    model=llm,
                    options=cli_args.llm_options,
                    prompt=prompt_txt.format(data=query_results, question=query),
                )
                answer = response['response']
                elapsed = round(time() - ts, 1)
                print(
                    json.dumps(
                        {
                            'query': query.strip(),
                            'sequence_no': n,
                            'prompt_id': prompt_id,
                            'answer': answer,
                            'elapsed': elapsed,
                            'collection': collection,
                            'embedding_model': embedding_model,
                            'search_results': cli_args.search_results,
                            'llm': llm,
                            'options': cli_args.llm_options,
                        }
                    )
                )



"""Handler for "coll" REPL command."""

from __future__ import annotations

from argparse import Namespace
from typing import Any

from rich.box import HORIZONTALS
from rich.table import Table

from .__common__ import ReplCommand, ReplCommandError, Style


# ------------------------------------------------------------------------------
@ReplCommand.register('coll', 'c')
class SetCollection(ReplCommand):
    """Set the database collection to be used."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument('collection', nargs='*')

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return (
            fr'c\[oll] [[{Style.arg}]COLLECTION ...]',
            'Show the collections in use or set to the ones specified.',
        )

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Set the LLM model."""

        def highlight_if(active: bool, v: Any) -> str:
            """Highlight a value using rich markup if it's active."""
            return f'[{Style.active}]{v}[/{Style.active}]' if active else str(v)

        collection_info: dict[str, dict[str, Any]] = self.repl_context.collection_info

        if args.collection:
            # Validate that we have the collections being requested.
            requested_collections = set(args.collection)
            available_collections = collection_info.keys()
            if unavaible_collections := requested_collections - available_collections:
                raise ReplCommandError(
                    f'Unavailable collections: {", ".join(sorted(unavaible_collections))}'
                )
            self.repl_context.collections = sorted(requested_collections)
            self.print(f'Collection(s) set to {", ".join(self.repl_context.collections)}')
            return True

        # Just show which collections are available in a table.
        columns = set()
        for c_info in collection_info.values():
            columns |= c_info.keys()
        # There are some columns we don't want to show
        columns = sorted(columns - {'files'})

        current_collections = set(getattr(self.repl_context, 'collections', []))
        table = Table(box=HORIZONTALS)
        table.add_column()  # Current collection indicator
        table.add_column('Name')
        for col in columns:
            table.add_column(col.replace('_', '\n').title())
        for c_name, c_info in sorted(collection_info.items()):
            active_row = c_name in current_collections
            table.add_row(
                highlight_if(active_row, '*') if active_row else '',
                highlight_if(active_row, c_name),
                *(highlight_if(active_row, c_info.get(k, '')) for k in columns),
            )

        self.print(table)
        return True



"""Model related utilities."""

from __future__ import annotations

from langchain_ollama import OllamaEmbeddings


# Map LLM control options to their type.
LLM_OPTIONS = {
    'mirostat': int,
    'mirostat_eta': float,
    'mirostat_tau': float,
    'num_ctx': int,
    'num_predict': int,
    'repeat_last_n': int,
    'repeat_penalty': float,
    'temperature': float,
    'seed': int,
    'stop': str,
    'tfs_z': float,
    'top_k': int,
    'top_p': float,
}


# ------------------------------------------------------------------------------
def get_embedding_dimensions(model: str) -> int:
    """Get the embedding dimensions of a model."""

    return len(OllamaEmbeddings(model=model).embed_documents(['Hello world'])[0])



"""Ollex exceptions."""

from __future__ import annotations


class OllexError(Exception):
    """Base class for ollex errors."""

    pass


class OllexInternalError(OllexError):
    """Internal error."""

    def __str__(self):
        """Get string representation."""

        return f'Internal error: {", ".join(str(s) for s in self.args)}'



"""Handler for "unset" REPL command."""

from __future__ import annotations

from argparse import Namespace
from contextlib import suppress

from rich.box import HORIZONTALS
from rich.table import Table

from ollex.lib.model import LLM_OPTIONS
from .__common__ import ReplCommand, Style


# ------------------------------------------------------------------------------
@ReplCommand.register('unset', 'u')
class UnsetLlm(ReplCommand):
    """Show / unset LLM options."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument(dest='options', nargs='*')

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return (
            fr'u\[nset] [[{Style.arg}]OPTION[/{Style.arg}] ...]',
            'Show / unset the current (non-default) LLM options (* = clear all).',
        )

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Unset LLM model options."""

        for option in args.options:
            match option:
                case '*':
                    self.repl_context.llm_options = {}
                case _:
                    with suppress(KeyError):
                        del self.repl_context.llm_options[option]

        table = Table(box=HORIZONTALS)
        table.add_column('Option')
        table.add_column('Value', justify='right')
        for option in sorted(LLM_OPTIONS):
            table.add_row(option, str(self.repl_context.llm_options.get(option, '')))

        self.print(table)

        return True



"""Handler for "set" REPL command."""

from __future__ import annotations

from argparse import Namespace
from readline import get_current_history_length, get_history_item

from rich.table import Table

from .__common__ import ReplCommand


# ------------------------------------------------------------------------------
@ReplCommand.register('history', 'hi', history=False)
class CommandHistory(ReplCommand):
    """Show command history.."""

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return r'hi\[story]', 'Show command history.'

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Set LLM model options."""

        table = Table(box=None, show_header=False)
        table.add_column('#', justify='right')
        table.add_column('Command')
        for n in range(1, get_current_history_length() + 1):
            table.add_row(str(n), get_history_item(n))
        self.print(table)

        return True



"""Handler for "prompt" REPL command."""

from __future__ import annotations

from argparse import Namespace

from .__common__ import ReplCommand, ReplCommandError, Style


# ------------------------------------------------------------------------------
@ReplCommand.register('prompt', 'p')
class Prompt(ReplCommand):
    """Show info on prompts."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument('prompt_id', nargs='?')

    # --------------------------------------------------------------------------
    @staticmethod
    def help() -> tuple[str, str]:
        """Command help."""

        return (
            fr'p\[rompt] [[{Style.arg}]ID]',
            'Show the prompt IDs in use or show the content of the one specified.',
        )

    # --------------------------------------------------------------------------
    def execute(self, args: Namespace) -> bool:
        """Show info on prompts."""

        if args.prompt_id:
            try:
                self.print(f'[{Style.prompt}]{self.repl_context.prompts[args.prompt_id]}')
            except KeyError:
                raise ReplCommandError('Invalid prompt ID')
            return True

        self.print('[bold]Prompts are:')
        for prompt_id in self.repl_context.prompts:
            self.print(f'    {prompt_id}')
        return True



"""Ollex CLI commands."""

__author__ = 'Murray Andrews'

import pkgutil
from importlib import import_module

from .__common__ import CliCommand as CliCommand

# Auto import our handler classes
for _, module_name, _ in pkgutil.iter_modules(__path__):
    if module_name.startswith('_'):
        continue
    import_module(f'.{module_name}', package=__name__)



"""Common components for docma CLI commands."""

from __future__ import annotations

from abc import ABC, abstractmethod
from argparse import Namespace
from collections.abc import Callable

from ollex.exceptions import OllexInternalError

__author__ = 'Murray Andrews'


# ------------------------------------------------------------------------------
class CliCommand(ABC):
    """ClI command handler."""

    commands: dict[str, type[CliCommand]] = {}
    name = None  # Set by @register decorator for subclasses.
    help_ = None  # Set by @register from first line of docstring.
    aliases = None  # Set by @register decorator for subclasses.

    # --------------------------------------------------------------------------
    @classmethod
    def register(cls, name: str, *aliases) -> Callable:
        """
        Register a CLI command handler class.

        This is a decorator. Usage is:

        .. code-block:: python

            @CliCommand.register('my_command', [aliases...])
            class MyCommand(CliCommand):
                ...

        The help for the command is taken from the first line of the docstring.
        """

        def decorate(cmd: type[CliCommand]):
            """Register the command handler class."""
            cmd.name = name
            cmd.aliases = aliases or []
            try:
                cmd.help_ = cmd.__doc__.splitlines()[0]
            except (AttributeError, IndexError):
                raise OllexInternalError(f'Class {cmd.__name__} must have a docstring')
            cls.commands[name] = cmd
            return cmd

        return decorate

    # --------------------------------------------------------------------------
    def __init__(self, subparser):
        """Initialize the command handler."""
        self.argp = subparser.add_parser(
            self.name, aliases=self.aliases, help=self.help_, description=self.help_
        )
        self.argp.set_defaults(handler=self)

    # --------------------------------------------------------------------------
    def add_arguments(self):  # noqa B027
        """Add arguments to the command handler."""
        pass

    # --------------------------------------------------------------------------
    @staticmethod  # noqa B027
    def check_arguments(args: Namespace):
        """
        Validate arguments.

        :param args:        The namespace containing the arguments.
        :raise ValueError:  If the arguments are invalid.
        """

        pass

    # --------------------------------------------------------------------------
    @staticmethod
    @abstractmethod
    def execute(args: Namespace) -> int:
        """
        Execute the CLI command with the specified arguments.

        :return:    Exit status.
        """
        raise NotImplementedError('execute')



"""Ollex "db" command."""

from argparse import Namespace
from hashlib import sha256
from pathlib import Path

import chromadb
import yaml
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_chroma import Chroma
from langchain_core.documents import Document  # noqa
from langchain_ollama import OllamaEmbeddings
from langchain_text_splitters import MarkdownHeaderTextSplitter
from rich.progress import Progress

from ollex.commands import CliCommand
from ollex.exceptions import OllexError
from ollex.lib.argparse import StoreNameValuePair
from ollex.lib.misc import chunk_list
from ollex.lib.model import get_embedding_dimensions

DEFAULT_COLLECTION = 'ai-boondoggle'

HEADERS_TO_SPLIT_ON = [
    ('#', 'Header 1'),
    ('##', 'Header 2'),
    ('###', 'Header 3'),
    ('####', 'Header 4'),
]

# Default values for embedding documents into a database collection.
# These can be set when a collection is created and cannot then be changed.
FROZEN_CONFIG = {
    'embedding_model': 'nomic-embed-text',
    'chunk_size': 1000,
    'chunk_overlap': 100,
}


# ------------------------------------------------------------------------------
@CliCommand.register('db')
class Db(CliCommand):
    """Load Markdown documents into a Chroma vector DB."""

    # --------------------------------------------------------------------------
    def add_arguments(self) -> None:
        """Add arguments to the command handler."""

        self.argp.add_argument('db_directory', help='Directory containing the Chroma DB.')
        self.argp.add_argument(
            'files',
            metavar='FILE.md',
            nargs='+',
            help='Markdown documents to add.',
        )

        self.argp.add_argument(
            '-c',
            '--collection',
            action='store',
            default=DEFAULT_COLLECTION,
            help=(
                'Database collection to which documents will be added. This will be'
                f' created if it does not exist. Default is {DEFAULT_COLLECTION}.'
            ),
        )

        self.argp.add_argument(
            '-p',
            '--param',
            metavar='NAME=VALUE',
            default={},
            action=StoreNameValuePair,
            help=(
                'Set the specified embedding control parameter(s). Default values are:'
                f' {", ".join(f"{k}={v}" for k, v in sorted(FROZEN_CONFIG.items()))}.'
                f' These cannot be changed for an existing collection and it is an error'
                f' to attempt to do so.'
            ),
        )

    @staticmethod  # noqa B027
    def check_arguments(args: Namespace) -> None:
        """Validate arguments."""

        for k, v in args.param.items():
            # Coerce values to the same type as the default
            try:
                args.param[k] = type(FROZEN_CONFIG[k])(v)
            except KeyError:
                raise ValueError(f'Parameter {k}: Not valid')
            except (ValueError, TypeError) as e:
                raise ValueError(f'Parameter {k}: {e}')

    # --------------------------------------------------------------------------
    @staticmethod
    def execute(args: Namespace) -> int:
        """Execute the command."""

        # Read existing config from the DB directory if present otherwise set up a
        # config for the collection from defaults. We also backfill any defaults
        # that happen to be missing from the existing config.
        config_path = Path(args.db_directory) / 'config.yaml'
        if config_path.exists():
            config = yaml.safe_load(config_path.read_text())
            config.setdefault('collections', {})
        else:
            config = {'collections': {}}

        collection_config = config['collections'].setdefault(args.collection, {})

        # Command line params are not allowed to override the config file.
        for k, v in args.param.items():
            if k in collection_config and collection_config[k] != v:
                raise OllexError(
                    f'{k} has already been set to {collection_config[k]} - cannot change now'
                )
        # Backfill any missing config values from our defaults
        collection_config = {'files': []} | FROZEN_CONFIG | args.param | collection_config
        config['collections'][args.collection] = collection_config

        md_splitter = MarkdownHeaderTextSplitter(
            headers_to_split_on=HEADERS_TO_SPLIT_ON, strip_headers=False
        )
        # Second splitter for large Markdown sections
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=collection_config['chunk_size'],
            chunk_overlap=collection_config['chunk_overlap'],
            length_function=len,
        )

        db_client = chromadb.PersistentClient(path=args.db_directory)
        db_client.get_or_create_collection(args.collection)
        embedding_fn = OllamaEmbeddings(model=collection_config['embedding_model'])
        vector_store = Chroma(
            client=db_client,
            collection_name=args.collection,
            embedding_function=embedding_fn,
        )
        collection_config['embedding_dimensions'] = get_embedding_dimensions(
            collection_config['embedding_model']
        )

        with Progress() as progress:
            if len(collection_config['files']) > 1:
                all_files_task = progress.add_task('Files', total=len(args.files))
            for filename in args.files:
                # We compare pre-loaded files with new files by hash to avoid duplication.
                file_content = Path(filename).read_text()
                file_hash = sha256(file_content.encode()).hexdigest()
                if file_hash in {f.get('sha256') for f in collection_config['files']}:
                    print(f'{filename}: Already added - skipping')
                    continue
                collection_config['files'].append({'name': filename, 'sha256': file_hash})

                md_split_docs = md_splitter.split_text(file_content)
                # Process each document - further split if needed
                split_docs = []
                for doc in md_split_docs:
                    if len(doc.page_content) > collection_config['chunk_size']:
                        for chunk in text_splitter.split_text(doc.page_content):
                            split_docs.append(
                                Document(page_content=chunk, metadata=doc.metadata.copy())
                            )
                    else:
                        split_docs.append(doc)
                # add meta data
                file_task = progress.add_task(filename, total=len(split_docs))
                for batch in chunk_list(split_docs, 10):
                    vector_store.add_documents(batch)
                    progress.update(file_task, advance=len(batch))
                progress.remove_task(file_task)

            # Update config file after each file in case of failure in a later file
            config_path.write_text(yaml.dump(config))
            if len(collection_config['files']) > 1:
                progress.update(all_files_task, advance=1)

        return 0



"""File utilities."""

from hashlib import sha256

BUFSIZ = 65536


# ------------------------------------------------------------------------------
def sha256_file(filename: str) -> str:
    """Calculate the sha256 hash of a file and return the hex digest."""

    sha = sha256()
    with open(filename, 'rb') as fp:
        while True:
            if not (buf := fp.read(BUFSIZ)):
                break
            print(buf)
            sha.update(buf)

    return sha.hexdigest()



"""setup.py for ollex."""

from __future__ import annotations

from importlib import resources
from pathlib import Path

from setuptools import find_packages, setup

__version__ = (resources.files('ollex') / 'VERSION').read_text().strip()

REPO_URL = 'https://github.com/jin-gizmo/ollex'
REQUIRES_PYTHON = '>=3.11.0'


# ------------------------------------------------------------------------------
def find_cli_entry_points(*cli_pkg: str, entry_point: str = 'main') -> list[str]:
    """Find CLI entry point scripts in the specified CLI packages."""

    entry_points = []
    for pkg in cli_pkg:
        pkg_path = Path(pkg.replace('.', '/'))
        if not pkg_path.is_dir():
            continue
        entry_points.extend(
            [
                f'{f.stem.replace("_", "-")}={pkg}.{f.stem}:{entry_point}'
                for f in pkg_path.glob('*.py')
                if not f.name.startswith('_')
            ]
        )
    return entry_points


# ------------------------------------------------------------------------------
# Import README.md and use it as the long-description. Must be in MANIFEST.in
with open('PYPI.md') as fp:
    long_description = '\n' + fp.read()

# ------------------------------------------------------------------------------
# Get pre-requisites from requirements.txt. Must be in MANIFEST.in
with open('requirements.txt') as fp:
    required = [s.strip() for s in fp.readlines()]

# ------------------------------------------------------------------------------
packages = find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*'])
setup(
    name='ollex',
    version=__version__,
    packages=packages,
    entry_points={
        'console_scripts': find_cli_entry_points(*(p for p in packages if p.endswith('.cli')))
    },
    url=REPO_URL,
    license='3-Clause BSD License',
    author='Murray Andrews',
    description='Open LLM Experimental Workbench',
    long_description=long_description,
    long_description_content_type='text/markdown',
    platforms=['macOS', 'Linux'],
    python_requires=REQUIRES_PYTHON,
    install_requires=required,
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



"""Argument parsing."""

from __future__ import annotations

from argparse import Action, ArgumentError, ArgumentParser
from typing import Any


# ------------------------------------------------------------------------------
class StoreNameValuePair(Action):
    """
    Used with argparse to store values from options of the form ``--option name=value``.

    The destination (self.dest) will be created as a dict {name: value}. This
    allows multiple name-value pairs to be set for the same option.

    Usage is:

    ::

        argparser.add_argument('-x', metavar='key=value', action=StoreNameValuePair)

    or

    ::

        argparser.add_argument('-x', metavar='key=value ...', action=StoreNameValuePair, nargs='+')

    """

    # --------------------------------------------------------------------------
    # noinspection PyUnresolvedReferences
    def __call__(self, parser, namespace, values, option_string=None):
        """Handle name=value option."""

        if not hasattr(namespace, self.dest) or not getattr(namespace, self.dest):
            setattr(namespace, self.dest, {})
        argdict = getattr(namespace, self.dest)

        if not isinstance(values, list):
            values = [values]
        for val in values:
            try:
                n, v = val.split('=', 1)
            except ValueError:
                raise ArgumentError(self, f'Expected "key=value", got "{val}"')
            argdict[n] = v


# ------------------------------------------------------------------------------
class ReplArgparserExitError(Exception):
    """When a premature exit from argparse is suppressed."""

    pass


class ReplArgparser(ArgumentParser):
    """Argparser that throws exception on bad arg instead of exiting."""

    def __init__(self, *args: Any, error_fmt: str = None, **kwargs: Any) -> None:
        """Add support for a custom argument for formatting errors."""
        super().__init__(*args, **kwargs)
        self.error_fmt = error_fmt or '{}'

    def exit(self, status=0, message=None):  # noqa A003
        """Stop argparse from exiting on bad options."""

        if status:
            raise ReplArgparserExitError(message)

    def error(self, message):
        """Raise an exception on bad arg instead of exiting."""
        raise ReplArgparserExitError(self.error_fmt.format(message))
